/*****************************************************************************
 * Media Library
 *****************************************************************************
 * Copyright (C) 2015-2018 Hugo Beauzée-Luyssen, Videolabs, VideoLAN
 *
 * Authors: Hugo Beauzée-Luyssen<hugo@beauzee.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#pragma once

#include "database/DatabaseHelpers.h"

namespace medialibrary
{

class Thumbnail;

class Thumbnail : public DatabaseHelpers<Thumbnail>
{
public:
    struct Table
    {
        static const std::string Name;
        static const std::string PrimaryKeyColumn;
        static int64_t Thumbnail::*const PrimaryKey;
    };

    enum class Origin : uint8_t
    {
        Artist,
        AlbumArtist,
        Album,
        Media,
        UserProvided,
    };

    Thumbnail( MediaLibraryPtr ml, sqlite::Row& row );
    Thumbnail( MediaLibraryPtr ml, std::string mrl, Origin origin, bool isGenerated );
    /**
     * @brief Thumbnail Constructor for a failure record
     */
    Thumbnail( MediaLibraryPtr ml );

    int64_t id() const;
    const std::string& mrl() const;
    bool update( std::string mrl, Origin origin );
    /**
     * @brief isValid Returns true if the thumbnail was correctly generated
     *
     * This only only relevent for thumbnail that are generated by the medialibrary
     */
    bool isValid() const;
    Origin origin() const;

    /**
     * @brief setMrlFromPrimaryKey Helper to set the thumbnail mrl based on a
     * thumbnail ID.
     * @param thumbnail The cached thumbnail entity. The value can be uncached
     * in which case it will be fetched and cached.
     * @param thumbnailId The thumbnail primary key
     */
    static bool setMrlFromPrimaryKey( MediaLibraryPtr ml,
                                      std::shared_ptr<Thumbnail>& thumbnail,
                                      int64_t thumbnailId, std::string mrl,
                                      Origin origin );

    static void createTable( sqlite::Connection* dbConnection );
    static std::shared_ptr<Thumbnail> create( MediaLibraryPtr ml, std::string mrl,
                                              Origin origin, bool isGenerated );
    /**
     * @brief createForFailure Creates a thumbnail record that denotes a failure
     * @param ml A medialibrary instance pointer
     *
     * This creates a record in the thumbnail table that denotes that a thumbnail
     * generation failed. A valid thumbnail will be generated, and can be updated
     * if required, or the record can be removed at a later time to force the regeneration
     */
    static std::shared_ptr<Thumbnail> createForFailure( MediaLibraryPtr ml );

    /**
     * @brief deleteFailureRecords Allow the thumbnail to retry any previously failed attempt
     *
     * This will delete all failure records
     */
    static bool deleteFailureRecords( MediaLibraryPtr ml );

    static const std::string EmptyMrl;

private:
    MediaLibraryPtr m_ml;
    int64_t m_id;
    std::string m_mrl;
    Origin m_origin;
    bool m_isGenerated;

    friend Thumbnail::Table;
};

}
